import abc
from typing import Any, Union

from zhtools.exceptions import ModuleRequired


class XlsxExporterInterface(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def export(self,
               header: list[str],
               data: list[list[Any]],
               path: str) -> None:
        """
        :param header: excel headers
        :param data: exported data
        :param path: file path

        >>> h = ['ColA', 'ColB', 'ColC']
        >>> d = [[1, 2, 3], [2, 2, 3]]
        >>> XlsxExporterInterface().export(h, d, '/tmp/test.xlsx')
        """
        pass

    @abc.abstractmethod
    def export_dict(self,
                    header: dict[str, str],
                    data: list[dict[str, Any]],
                    path: str) -> None:
        """
        :param header: excel headers
        :param data: exported data
        :param path: file path

        >>> h = {'col1': 'ColA', 'col2': 'ColB', 'col3': 'ColC'}
        >>> d = [{'col1': 1, 'col2': 2, 'col3': 3}, {'col1': 2, 'col2': 2, 'col3': 4}]
        >>> XlsxExporterInterface().export_dict(h, d, '/tmp/test.xlsx')
        """
        pass


class XlsxExporter(XlsxExporterInterface):

    def _export(self, writer, header, data, path: str, sheetname: str = None):
        try:
            import openpyxl
        except ImportError:
            raise ModuleRequired('openpyxl', '3.0.6')

        wb = openpyxl.Workbook()
        ws = wb.active
        if sheetname:
            ws.title = sheetname

        writer(ws, header, data)
        wb.save(path)

    def export(self,
               header: list[str],
               data: list[list[Union[str, int, float]]],
               path: str,
               sheetname: str = None,
               ):
        self._export(self.write_sheet, header, data, path, sheetname)

    def export_dict(self,
                    header: dict[str, str],
                    data: list[dict[str, Union[str, int, float]]],
                    path: str,
                    sheetname: str = None):
        self._export(self.write_sheet2, header, data, path, sheetname)

    @staticmethod
    def write_sheet(ws,
                    header: list[str],
                    data: list[list[Union[str, int, float]]]):
        ws.append(header)
        for row in data:
            ws.append(row)

    @staticmethod
    def write_sheet2(ws,
                     header: dict[str, str],
                     data: list[dict[str, Union[str, int, float]]]):
        ws.append(list(header.values()))
        for item in data:
            ws.append([item.get(k, '') for k in header])



import string
import uuid


def uuid4_hex(duplicate: int = 1) -> str:
    """
    Generate a uuid
    :param duplicate: duplicate times
    :return: uuid string
    """
    s = ''
    for _ in range(duplicate):
        s += uuid.uuid4().hex
    return s


def short_uuid() -> str:
    """
    Generate a short uuid
    :return:
    """
    uuid_ = uuid4_hex()
    chars = string.ascii_lowercase + string.digits + string.ascii_uppercase
    s = ''
    for i in range(0, 8):
        sub = uuid_[i * 4:i * 4 + 4]
        x = int(sub, 16)
        s += chars[x % 0x3E]
    return s



import abc
import os
import pickle
import time
import shelve
from typing import Any, AnyStr, Optional


class CacheStorageInterface(metaclass=abc.ABCMeta):

    @abc.abstractmethod
    def get(self, key: AnyStr) -> Any:
        pass

    @abc.abstractmethod
    def set(self,
            key: AnyStr,
            value: AnyStr,
            expire: Optional[int] = None) -> None:
        pass

    @abc.abstractmethod
    def delete(self, key: AnyStr) -> None:
        pass


class MemoryCacheStorage(CacheStorageInterface):
    """Cache storage use memory"""

    def __init__(self):
        self.__storage = {}
        self.__expires = {}

    def _get(self, key: AnyStr) -> Any:
        return self.__storage.get(key)

    def _set(self,
             key: AnyStr,
             value: AnyStr,
             expire: Optional[int] = None) -> None:
        self.__storage[key] = value

    def _delete(self, key: AnyStr) -> None:
        self.__storage.pop(key, None)

    def get(self, key: AnyStr) -> Any:
        expire = self.__expires.get(key)
        if expire and expire >= time.time():
            self.delete(key)
            return None
        return self._get(key)

    def set(self, key: AnyStr, value: AnyStr,
            expire: Optional[int] = None) -> None:
        self._set(key, value, expire)
        if expire is not None:
            self.__expires[key] = time.time() + expire

    def delete(self, key: AnyStr) -> None:
        self._delete(key)
        self.__expires.pop(key, None)


class FileCacheStorage(MemoryCacheStorage):
    """Cache storage use file"""

    def __init__(self, path: str = '/tmp'):
        super().__init__()
        self.path = path

    def _get(self, key: AnyStr) -> Any:
        with open(os.path.join(self.path, str(key)), 'rb') as f:
            return pickle.load(f)

    def _set(self,
             key: AnyStr,
             value: Any,
             expire: Optional[int] = None) -> None:
        with open(os.path.join(self.path, str(key)), 'wb') as f:
            pickle.dump(value, f)

    def _delete(self, key: AnyStr) -> None:
        os.remove(os.path.join(self.path, key))


class ShelveCacheStorage(MemoryCacheStorage):

    def __init__(self, db: str = 'shelve_cache'):
        super().__init__()
        self.db = db

    def _set(self,
             key: AnyStr,
             value: AnyStr,
             expire: Optional[int] = None) -> None:
        with shelve.open(self.db) as db:
            db[key] = value

    def _get(self, key: AnyStr) -> Any:
        with shelve.open(self.db) as db:
            return db.get(key)

    def _delete(self, key: AnyStr) -> None:
        with shelve.open(self.db) as db:
            db.pop(key, None)



from unittest import TestCase

from zhtools.cache import cache


called_times = 0


@cache(key_prefix='test')
def simple_cache():
    global called_times
    called_times += 1
    return called_times


called = False


@cache(key_prefix='test1')
def cache_with_params(a: int, b: int):
    global called
    called = True
    return a + b


called2 = False


@cache(key_prefix='test2')
def cache_with_ignore_exception(a, b):
    global called2
    called2 = True
    return a / b


called3 = False


@cache(key_prefix='test3', ignore_exception=False)
def cache_with_exception(a, b):
    global called3
    called3 = True
    return a / b


@cache(key_prefix='test4')
def cache1():
    return 1


class TestCache(TestCase):

    def test_use_cache(self):
        self.assertEqual(called_times, 0)
        ret = simple_cache()
        self.assertEqual(called_times, 1)
        self.assertEqual(ret, 1)
        simple_cache()
        self.assertEqual(called_times, 1)
        self.assertEqual(ret, 1)

    def test_use_cache_with_params(self):
        global called
        ret = cache_with_params(1, 2)
        self.assertEqual(ret, 3)
        self.assertTrue(called)

        called = False
        ret = cache_with_params(1, 2)
        self.assertEqual(ret, 3)
        self.assertFalse(called)

        ret = cache_with_params(2, 2)
        self.assertEqual(ret, 4)
        self.assertTrue(called)

    def test_cache_with_exception(self):
        global called2
        with self.assertRaises(ZeroDivisionError):
            cache_with_ignore_exception(2, 0)

        self.assertTrue(called2)

        called2 = False
        with self.assertRaises(ZeroDivisionError):
            cache_with_ignore_exception(2, 0)

        self.assertTrue(called2)

    def test_cache_with_exception_and_not_ignore(self):
        global called3
        with self.assertRaises(ZeroDivisionError):
            cache_with_exception(2, 0)

        self.assertTrue(called3)

        called3 = False
        with self.assertRaises(ZeroDivisionError):
            cache_with_exception(2, 0)

        self.assertFalse(called3)

    def test_update_cache_result(self):
        ret = cache1()
        self.assertEqual(ret, 1)
        ret = cache1()
        self.assertEqual(ret, 1)

        cache1.update_cached_result(2)
        ret = cache1()
        self.assertEqual(ret, 2)



import functools
import inspect
import pickle
from typing import Any, Callable, Optional, Union

from zhtools.cache.storages import CacheStorageInterface, MemoryCacheStorage
from zhtools.random import uuid4_hex


def default_get_key_suffix(*args, **kwargs):
    """Generate cache key suffix from function parameters."""
    r = {'args': args, 'kwargs': kwargs}
    return pickle.dumps(r)


def make_cache_key(key_prefix: str,
                   key_suffix: Union[Callable, str],
                   *args,
                   **kwargs) -> str:
    """Generate cache key from function parameters"""
    if inspect.isfunction(key_suffix):
        suffix = key_suffix(*args, **kwargs)
    else:
        suffix = key_suffix

    return f'{key_prefix}:{suffix}'


class Cache:
    """
    Cache tool
    >>> @Cache(key_prefix='cache_key')
    >>> def foo():
    >>>     return True

    storage: cache storage, default storage is use by memory.
    """
    storage: CacheStorageInterface = MemoryCacheStorage()

    def __init__(self,
                 key_prefix: str = '',
                 key_suffix: Union[Callable, str] = default_get_key_suffix,
                 cache_expires: Optional[int] = None,
                 ignore_exception: bool = True):
        """
        :param key_prefix: the prefix of cache key
        :param key_suffix: the suffix of cache key
        :param cache_expires: the expires of cache, seconds
        :param ignore_exception: while is True, it will skip cache when
                                 function raised exception
        """
        self.key_prefix = key_prefix
        self.key_suffix = key_suffix
        self.cache_expires = cache_expires
        self.name = uuid4_hex()
        self.ignore_exception = ignore_exception

    def make_cache_key(self, *args, **kwargs):
        return make_cache_key(self.key_prefix, self.key_suffix, *args, **kwargs)

    def get_cached_result(self, key: str) -> Any:
        r = self.storage.get(key)
        if not r:
            return

        if not isinstance(r, bytes):
            return r

        r = pickle.loads(r)
        if r:
            if isinstance(r, Exception):
                raise r

            return r

    def __call__(self, func: Callable):

        def wrapped(*args, **kwargs):
            key = self.make_cache_key(*args, **kwargs)
            result = self.get_cached_result(key)
            if not result:
                try:
                    result = func(*args, **kwargs)
                except Exception as e:
                    if self.ignore_exception:
                        raise e

                    result = e
                self.storage.set(key, result, self.cache_expires)

            if isinstance(result, Exception):
                raise result
            return result

        def update_cached_result(result: Any, *args, **kwargs) -> None:
            """
            method to update cache
            >>> @Cache()
            >>> def foo():
            >>>    return 1
            >>> foo.update_cached_result(0)
            >>> foo()
                0

            :param result: cache value
            :param args: args for making cache key
            :param kwargs: kwargs for making cache key
            """
            key = self.make_cache_key(*args, **kwargs)
            if result is None:
                self.storage.delete(key)
            else:
                self.storage.set(key, result, self.cache_expires)

        functools.update_wrapper(wrapped, func)
        wrapped.update_cached_result = update_cached_result
        return wrapped


cache = Cache



import logging
import os
import time
from pathlib import Path
from unittest import TestCase, skipIf

from tests.common import fake

skip = False
try:
    import openpyxl
    from zhtools.io_tools.readers.xlsx import XlsxReader
except ImportError:
    logging.warning('openpyxl not install, skip...')
    skip = True


@skipIf(skip, 'openpyxl not install.')
class TestXlsxReader(TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        cls.filename = Path('/tmp') / Path(str(time.time()) + '.xlsx')
        wb = openpyxl.Workbook()
        ws = wb.active

        cls.data = []
        for _ in range(4):
            row = [fake.pyint() for _ in range(3)]
            cls.data.append(row)
            ws.append(row)

        wb.save(cls.filename)
        wb.close()

    @classmethod
    def tearDownClass(cls) -> None:
        os.remove(cls.filename)

    def test_read(self):
        with XlsxReader(self.filename) as f:
            data = f.read()

        self.assertEqual(data, self.data)

    def test_readlines(self):
        with XlsxReader(self.filename) as f:
            data = f.readlines()

        self.assertEqual(data, list(map(tuple, self.data)))

    def test_read_col(self):
        with XlsxReader(self.filename) as f:
            data = list(f.read_col(0))

        self.assertEqual(data, [i[0] for i in self.data])



class ModuleRequired(Exception):

    def __init__(self, module: str, version: str = None):
        self.module = module
        self.version = version

    def __str__(self) -> str:
        module = self.module
        if self.version:
            module += f'>={self.version}'
        return (f'Module [{module}]'
                f' is required. Please use pip to install it.')



import datetime


datetime_format = '%Y-%m-%d %H:%M:%S'
date_format = '%Y-%m-%d'
compact_datetime_format = '%Y%m%d%H%M%S'
compact_date_format = '%Y%m%d'


def local_datetime(days: int = None):
    dt = datetime.datetime.now()
    td = None
    if days:
        td = datetime.timedelta(days=days)
    if td:
        dt += td
    return dt


def timestamp_to_datetime(ts: int,
                          tz: datetime.tzinfo = None) -> datetime.datetime:
    return datetime.datetime.fromtimestamp(ts, tz=tz)


def datetime_to_timestamp(dt: datetime.datetime) -> int:
    return int(dt.timestamp())


def date_to_datetime(dt: datetime.date) -> datetime.datetime:
    return datetime.datetime(year=dt.year,
                             month=dt.month,
                             day=dt.day)


def date_from_string(date_str: str, format: str = date_format) -> datetime.date:
    return datetime.datetime.strptime(date_str, format).date()


start_of_date = date_to_datetime


def end_of_date(dt: datetime.date) -> datetime.datetime:
    return date_to_datetime(dt) + datetime.timedelta(
        days=1) - datetime.timedelta(seconds=1)



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import Any, Union

from zhtools.convertors import underline_to_camel_case

try:
    from pydantic import BaseModel
except ImportError:
    from zhtools.exceptions import ModuleRequired
    raise ModuleRequired('pydantic')


@dataclass
class Model:
    name: str
    children: dict[str, Union[type, Model]]

    def __eq__(self, other):
        return self.name == other.name and self.children == other.children

    def _render(self) -> list[str]:
        ret = [f'class {underline_to_camel_case(self.name)}(BaseModel):']
        for n, m in self.children.items():
            if isinstance(m, Model):
                t = m.render_type()
                rendered = m._render()
                if rendered:
                    ret = rendered + ['', ''] + ret
            else:
                t = m.__name__

            ret.append(f'    {n}: {t}')

        return ret

    def render_type(self) -> str:
        return underline_to_camel_case(self.name)

    def render_child_model(self) -> list[str]:
        ret = []
        for n, m in self.children.items():
            if isinstance(m, Model):
                rendered = m._render()
                if rendered:
                    ret = rendered + ['', ''] + ret

        return ret

    def render(self) -> str:
        return '\n'.join(self._render())


class ModelArray(Model):

    def __init__(self, name: str, child: Union[type, Model]):
        super().__init__(name, {})
        self.child = child

    def __eq__(self, other):
        return self.name == other.name and self.child == other.child

    @property
    def child_type(self) -> str:
        if isinstance(self.child, Model):
            return underline_to_camel_case(self.child.name)
        else:
            return self.child.__name__

    def _render(self) -> list[str]:
        if isinstance(self.child, Model):
            return self.child._render()
        return []

    def render_type(self) -> str:
        return f'list[{self.child_type}]'

    def render_child_model(self) -> list[str]:
        if isinstance(self.child, Model):
            return self.child._render() + ['', '']
        return []


class Json2Model:

    @staticmethod
    def parse(data: str, model_name: str = 'model') -> str:
        data = json.loads(data)
        if not isinstance(data, dict):
            raise TypeError()
        return Json2Model.parse_dict(model_name, data).render()

    @staticmethod
    def parse_dict(model_name: str, data: dict) -> Model:

        def _parse(name: str, v: Any) -> Union[type, Model]:
            if isinstance(v, list):
                return ModelArray(name, _parse(name, v[0]))
            elif isinstance(v, dict):
                return Json2Model.parse_dict(name, v)
            else:
                return type(v)

        m = Model(model_name, {})
        for k, v in data.items():
            m.children[k] = _parse(k, v)

        return m




import functools
from contextlib import contextmanager
from inspect import isfunction
from typing import Callable, Optional, Union

from zhtools import ignore_exception
from zhtools.exceptions import ModuleRequired

_redis_pool = None
_redis_client = None


def init(redis_client=None, redis_pool=None):
    global _redis_pool
    global _redis_client
    _redis_client = redis_client
    _redis_pool = redis_pool


def get_client():
    if _redis_client:
        return _redis_client
    if _redis_pool:
        try:
            import redis
            return redis.Redis(connection_pool=_redis_client)
        except ImportError:
            raise ModuleRequired('redis')


@contextmanager
def concurrent_lock(key: str,
                    ttl: int = 3,
                    blocking_timeout: Optional[int] = None,
                    redis_cli=None):
    """
    >>> with concurrent_lock(key='lock') as l:
    >>>     if not l:
    >>>         raise Exception()
    >>>     ...
    """
    redis_cli = redis_cli or get_client()
    lock = redis_cli.lock(key,
                          timeout=ttl,
                          blocking_timeout=blocking_timeout)
    locked = lock.acquire()

    try:
        yield locked
    finally:
        if locked:
            with ignore_exception():
                lock.release()


class GetLockError(Exception):
    pass


def concurrent_limit(key: Union[str, Callable],
                     ttl: int = 3,
                     blocking_timeout: Optional[int] = None):
    """
    Use directly
    >>> @concurrent_limit
    >>> def foo(k):
    >>>     pass

    Use with parameters
    >>> @concurrent_limit(key='lock')
    >>> def foo():
    >>>     pass
    """
    def decorator(func):
        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            nonlocal key
            if isfunction(key):
                key = key.__name__

            with concurrent_lock(key, ttl, blocking_timeout) as lock:
                if not lock:
                    raise GetLockError()
                return func(*args, **kwargs)

        return wrapped

    if isfunction(key):
        return decorator(key)

    return decorator



import functools
import logging
from contextlib import contextmanager
from typing import Callable

from zhtools.type_hint import AnyNumber


def property_with_cache(func: Callable):
    """
    Add an underscore before the property name for caching.
    >>> class Foo
    >>>     @property_with_cache
    >>>     def prop1(self):
    >>>         return True

    >>> foo = Foo()
    >>> foo.prop1
    True
    >>> getattr(foo, '_prop1')
    True
    """
    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        instance = args[0]
        property_name = func.__name__
        private_prop_name = '_' + property_name
        if not getattr(instance, private_prop_name, False):
            setattr(instance, private_prop_name, func(*args, **kwargs))
        return getattr(instance, private_prop_name, None)

    return property(wrapped)


@contextmanager
def ignore_exception(logger: logging.Logger = None):
    try:
        yield
    except:
        if logger:
            logger.exception('Exception raise but has been ignored.')


def safe_divide(a: AnyNumber,
                b: AnyNumber,
                decimal_len: int = 2,
                while_zero: AnyNumber = 0) -> str:
    """
    Safe division avoiding ZeroDivisionError
    :param a: dividend
    :param b: divisor
    :param decimal_len: decimal digits
    :param while_zero:
    :return: division result
    """
    fm = f'%.{decimal_len}f'
    return fm % (a / b) if b else fm % while_zero



def underline_to_camel_case(v: str) -> str:
    return v.replace('_', ' ').title().replace(' ', '')


def underline_to_small_camel_case(v: str) -> str:
    v = underline_to_camel_case(v)
    return v[0].lower() + v[1:]


def camel_case_to_underline(v: str) -> str:
    ret = v[0].lower()
    for i in v[1:]:
        if i.isupper():
            ret += '_'

        ret += i.lower()

    return ret



from unittest import TestCase

from zhtools.code_generator.json2model import (Json2Model, Model,
                                               ModelArray)


class TestJson2Model(TestCase):

    def test_parse_dict(self):
        testcases = [
            ({"action": 230002}, Model('test', {'action': int})),
            ({"action": "230002"}, Model('test', {'action': str})),
            ({"action": [1, 2]}, Model('test', {'action': ModelArray('action', int)})),
            ({"action": [{'key': 0}]}, Model('test', {'action': ModelArray('action', Model('action', {'key': int}))})),
            ({"action": {"key1": 1, "key2": "test"}},
             Model('test', {'action': Model('action', {'key1': int, 'key2': str})})),
            ({'action': {'key': [{'key1': 1}]}},
             Model('test', children={
                 'action': Model('action', children={
                     'key': ModelArray('key', child=Model('key', children={
                         'key1': int
                     }))
                 })
             })
             )
        ]
        for param, result in testcases:
            ret = Json2Model.parse_dict('test', data=param)
            self.assertEqual(ret, result)

    def test_render(self):
        testcases = [
            (Model('test', {'action': int}), ['class Test(BaseModel):', '    action: int']),
            (Model('test', {'action': str}), ['class Test(BaseModel):', '    action: str']),
            (Model('test', {'action': ModelArray('action', int)}), ['class Test(BaseModel):', '    action: list[int]']),
            (Model('test', {'action': ModelArray('action', Model('action', {'key': int}))}),
             ['class Action(BaseModel):', '    key: int', '', '', 'class Test(BaseModel):', '    action: list[Action]']),
            (Model('test', {'action': Model('action', {'key1': int, 'key2': str})}),
             ['class Action(BaseModel):', '    key1: int', '    key2: str', '', '', 'class Test(BaseModel):', '    action: Action']),
            (Model('test', children={
                'action': Model('action', children={
                    'key': ModelArray('key', child=Model('key', children={
                        'key1': int
                    }))
                })
            }),
             ['class Key(BaseModel):', '    key1: int', '', '', 'class Action(BaseModel):', '    key: list[Key]', '', '', 'class Test(BaseModel):', '    action: Action']
             )

        ]
        for model, result in testcases:
            ret = model._render()
            self.assertEqual(ret, result)

    def test_parse(self):
        ret = Json2Model.parse('{"action": 230002}')
        self.assertEqual(ret, 'class Model(BaseModel):\n    action: int')



from collections import Generator
from pathlib import Path
from typing import Any, Callable, Optional

from zhtools.exceptions import ModuleRequired
from .interfaces import ReaderInterface

try:
    import openpyxl
except ImportError:
    raise ModuleRequired('openpyxl', '3.0.6')


class XlsxReader(ReaderInterface):

    def __init__(self, path: Path, sheet_name: str = None, sheet_idx: int = None):
        super().__init__(path)
        self.sheet_name = sheet_name
        self.sheet_idx = sheet_idx
        self.wb = openpyxl.open(self.path)

    def get_sheet_name(self) -> Optional[str]:
        if self.sheet_name:
            return self.sheet_name

        if not self.sheet_idx:
            return

        names = self.wb.get_sheet_names()
        return names[self.sheet_idx]

    def get_sheet(self):
        sheet_name = self.get_sheet_name()
        if sheet_name:
            sheet = self.wb.get_sheet_by_name(sheet_name)
        else:
            sheet = self.wb.active
        return sheet

    def close(self):
        self.wb.close()

    def read(self) -> list[list[Any]]:
        sheet = self.get_sheet()
        return [list(row) for row in sheet.iter_rows(values_only=True)]

    def readline(self) -> Generator[tuple, None, None]:
        sheet = self.get_sheet()
        return sheet.iter_rows(values_only=True)

    def readlines(self) -> list[Any]:
        return list(self.readline())

    def read_col(self,
                 idx: int = 0,
                 filter: Callable[[Any], bool] = None
                 ) -> Generator[Any, None, None]:
        sheet = self.get_sheet()
        for col in sheet.iter_cols(min_col=idx + 1,
                                   max_col=idx + 1,
                                   values_only=True):
            for val in col:
                if filter and not filter(val):
                    continue

                yield val



import setuptools


with open('README.md', 'r') as f:
    long_description = f.read()


setuptools.setup(
    name='zhtools',
    version='0.0.4',
    author='zhyipeng',
    author_email='zhyipeng@outlook.com',
    description='Some simple tool methods like cache, exporter and so on.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/zhyipeng/zhtools',
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ]
)




def read_file(filename: str, offset: int, size: int) -> str:
    """Read the specified interval content of text file"""
    with open(filename, 'r') as f:
        f.seek(offset)
        return f.read(size)



import abc
from collections import Generator
from pathlib import Path
from typing import Any


class ReaderInterface(metaclass=abc.ABCMeta):

    def __init__(self, path: Path):
        self.path = path

    @abc.abstractmethod
    def read(self) -> Any:
        pass

    @abc.abstractmethod
    def readlines(self) -> list[Any]:
        pass

    @abc.abstractmethod
    def readline(self) -> Generator[Any, None, None]:
        pass

    def close(self):
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()
        if exc_val:
            raise exc_val
